import numpy as np
import cupy as cp
import massfunc as mf 
from astropy.constants import m_p
from astropy import units as u
from scipy.optimize import brentq
from scipy.integrate import quad_vec,quad


cosmo = mf.SFRD()
m_H = (m_p.to(u.M_sun)).value #M_sun
omega_b = cosmo.omegab
omega_m = cosmo.omegam

def xim(m,z):
    A,B,C,D,E,F,G = 4.4 , 0.334, 0.023, 0.199, -0.042, 0,1
    M7 = m/1e7
    F0 = 1.0
    return 1+A*M7**(B+C*np.log10(M7))*(F+G*(1+z)/10)* F0**(D+E*np.log10(F0))

def fstar(Mh): # Donnan 25 
    eps0 = 0.16
    Mc = 10**11.7
    beta = 0.9
    gamma = 0.65
    return 2*eps0 * ((Mh / Mc)**-beta + (Mh / Mc)**gamma)**-1


def qion_sb99(Muv):
    if Muv > 20:
        return 1000
    elif Muv >= 16.5:
        return 4065000 * np.exp(-0.420 * Muv)
    elif Muv >= 13:
        return -32336 * np.log10(Muv) + 43720
    else:
        return 7700
    
#### GPU version of interp1d
def interp1d_gpu(x, y, x_new):
    idx = cp.searchsorted(x, x_new, side='right') - 1
    idx = cp.clip(idx, 0, x.size - 2)
    dx = x[idx + 1] - x[idx]
    dy = y[idx + 1] - y[idx]
    return y[idx] + dy * (x_new - x[idx]) / dx


from .powerspec import MassFunctions
import numpy as np
from scipy.integrate import simpson
from scipy.interpolate import interp1d
import astropy.units as u
from cosfunc import n_H,dtdz,H
import os
from scipy.integrate import quad_vec,quad
from astropy.constants import m_p



class Ion:

    def __init__(self,z,fesc=0.2,kakaka=0.7*1e-28,xi_ion=10**25.6,A2byA1=0.1,ktrans=200,alpha=2.0,beta=0.0):
        self.cosmo = MassFunctions(A2byA1=A2byA1,kMpc_trans=ktrans,alpha=alpha,beta=beta)
        self.A2byA1 = A2byA1
        self.kMpc_trans = ktrans
        self.alpha = alpha
        self.beta = beta
        self.z = z
        self.kakaka = kakaka
        self.xi_ion = xi_ion
        self.fesc = fesc
        self.deltaV_interp = np.linspace(-0.999, 2, 1000)  # delta_V
        self.M_min = self.cosmo.M_vir(0.61, 1e4, z)  # M_sun
        self.M_J = self.cosmo.M_Jeans(z)  # M_sun


    ### Source term
    def fstar(self, Mh):
        eps0 = 0.16
        Mc = 10**11.7
        beta = 0.9
        gamma = 0.65
        return 2*eps0 * ((Mh / Mc)**-beta + (Mh / Mc)**gamma)**-1

    def fduty(self, M):
        al = 1.5
        Mc = 6e7
        return (1 + (2.**(al / 3.) - 1) * (M / Mc)**-al)**(-3. / al)

    def dMdt(self, M, z):
        return 24.1 * (M / (1e12))**1.094 * (1 + 1.75 * z) * np.sqrt(self.cosmo.omegam * (1 + z)**3 + (1-self.cosmo.omegam))  # solmass/yr

    def rhosfrdiff(self, Mh, Mv,deltaV,z):
        diff = self.fstar(Mh) * self.cosmo.omegab / self.cosmo.omegam * self.dMdt(Mh, z) * self.cosmo.dndmeps(Mh,Mv,deltaV, z) * self.fduty(Mh)
        return diff
    
    def rhosfr(self,Mv,deltaV):
        try:
            Nxi_arr = np.load(f'.rhosfr_dz_Interp_init/NxiAtz{self.z:.2f}/Nxi_arr_Mv_{Mv:.3f}at_z={self.z:.2f}_A{self.A2byA1}_k{self.kMpc_trans}_alpha{self.alpha}_beta{self.beta}.npy')
        except FileNotFoundError:
            os.makedirs(f'.rhosfr_dz_Interp_init/NxiAtz{self.z:.2f}', exist_ok=True)
            M_min = self.cosmo.M_vir(0.61, 1e4, self.z)
            M_max = min(self.cosmo.M_vir(0.61, 1e8, self.z),0.9*Mv)
            Mh_interp = np.linspace(np.log10(M_min), np.log10(M_max), 1000)  # log10(M)
            deltaV_grid, m_grid = np.meshgrid(self.deltaV_interp, Mh_interp, indexing='ij')
            mh = self.rhosfrdiff(Mh=10**m_grid, Mv=Mv,deltaV=deltaV_grid, z=self.z)
            integrand = mh*10**m_grid * np.log(10)  # dN/dlogM -> dN/dM
            # Integrate using Simpson's rule
            Nxi_arr_save = simpson(integrand, x=Mh_interp, axis=1)
            np.save(f'.rhosfr_dz_Interp_init/NxiAtz{self.z:.2f}/Nxi_arr_Mv_{Mv:.3f}at_z={self.z:.2f}_A{self.A2byA1}_k{self.kMpc_trans}_alpha{self.alpha}_beta{self.beta}.npy', Nxi_arr_save)
            Nxi_arr = np.load(f'.rhosfr_dz_Interp_init/NxiAtz{self.z:.2f}/Nxi_arr_Mv_{Mv:.3f}at_z={self.z:.2f}_A{self.A2byA1}_k{self.kMpc_trans}_alpha{self.alpha}_beta{self.beta}.npy')
        Simpson_Interp = interp1d(self.deltaV_interp, Nxi_arr, kind='cubic')
        return Simpson_Interp(deltaV)

    def rho_sfr_st(self):
        M_min = self.cosmo.M_vir(0.61, 1e4, self.z)
        M_max = self.cosmo.M_vir(0.61, 1e8, self.z)
        Mh_interp = np.linspace(np.log10(M_min), np.log10(M_max), 1000) 
        def rhodiff(Mh):
            return self.fstar(Mh) * self.cosmo.omegab / self.cosmo.omegam * self.dMdt(Mh, self.z) * self.cosmo.dndmst(Mh,self.z) * self.fduty(Mh)
        mh = rhodiff(Mh=10**Mh_interp)
        integrand = mh*10**Mh_interp * np.log(10)  # dN/dlogM -> dN/dM
        return simpson(integrand, Mh_interp, axis=0)
    
    def rho_UV(self,Mv,deltaV):
        return self.rhosfr(Mv,deltaV=deltaV)/self.kakaka
    
    def nion_dot_st(self):
        return self.fesc * self.xi_ion * self.rho_sfr_st()/self.kakaka * u.s**-1 *u.Mpc**-3
    
    def nion_dot(self,Mv,deltaV):
        return self.fesc * self.xi_ion * self.rho_UV(Mv,deltaV) * u.s**-1 *u.Mpc**-3

    def nion_dz(self,Mv,deltaV):
        return (self.nion_dot(Mv,deltaV)/n_H(deltaV) * dtdz(self.z)).to(u.dimensionless_unscaled).value * (n_H(0).to(u.Mpc**-3)).value  # comoving

    def nion_dz_st(self):
        return (self.nion_dot_st()/n_H(0) * dtdz(self.z)).to(u.dimensionless_unscaled).value * (n_H(0).to(u.Mpc**-3)).value  # comoving
    
    ### IGM term
    def trec(self,deltaV):
        alpha_B = 2.59e-13  # cm^3/s at 10^4 K
        clumping_factor = 3.0
        Xe = 1.09
        return 1.0 / (clumping_factor * alpha_B * (1 + deltaV) * Xe*n_H(deltaV).to(1/u.cm**3).value * (1 + self.z)**3)  # s

    def n_HI(self,deltaV):
        return n_H(deltaV).to(u.Mpc**-3).value  ### coming number density of hydrogen atoms in cm^-3
    
    def CHII(self,z):
        return 2.9*((1+z)/6)**-1.1

    def dtdz(self,z):
        return ((-1/(H(z)*(1+z))).to(u.s)).value

    def dnrec_dz_path(self,deltaV,xHII_Field:np.ndarray)->np.ndarray:
        x_HE = 1.08
        # CIGM = self.CHII(z)
        CIGM = 3.0
        nh = self.n_HI(deltaV)*(u.Mpc**-3).to(u.cm**-3)
        Q_HII = xHII_Field
        # alpha_A = 4.2e-13 #cm**3/s
        alpha_B = 2.59e-13  # cm^3/s at 10^4 K
        differential_trans = self.dtdz(self.z)
        return -CIGM*x_HE*alpha_B*nh*Q_HII*(1+self.z)**3 * differential_trans
    
    ### Minihalo term
    def dndmeps(self,M,Mr,deltaV,z):
        deltaL = self.cosmo.delta_L(deltaV,self.z)
        sig1 = self.cosmo.sigma2_interp(M) - self.cosmo.sigma2_interp(Mr)
        del1 = self.cosmo.deltac(z) - deltaL
        return self.cosmo.rhom * (1 + deltaV) / M / np.sqrt(2 * np.pi) * abs(self.cosmo.dsigma2_dm_interp(M)) * del1 / (sig1**(3 / 2)) * np.exp(-del1 ** 2 / (2 * sig1))

    def dEPS_dz(self,M,Mv,deltaV,z):
        return (self.dndmeps(M,Mv,deltaV,z+0.001*z) - self.dndmeps(M,Mv,deltaV,z-0.001*z)) / (0.002*z)

    def Xim(self,m,z):
        A,B,C,D,E,F,G = 4.4 , 0.334, 0.023, 0.199, -0.042, 0,1
        M7 = m/1e7
        F0 = 1.0
        return 1+A*M7**(B+C*np.log10(M7))*(F+G*(1+z)/10)* F0**(D+E*np.log10(F0))
    
    def dNxi_dz(self,m,deltaR,Mv,z):
        m_H = (m_p.to(u.M_sun)).value #M_sun
        return self.Xim(m,z)*m*self.dEPS_dz(m,Mv,deltaR,z)/ m_H * self.cosmo.omegab / self.cosmo.omegam

    
    def Simpson_Nxi_dz(self,delta_R:np.ndarray,M_v:float,z:float) -> np.ndarray:
        try:
            Nxi_arr = np.load(f'.Nxi_dz_Interp_init/NxiAtz{self.z:.2f}/Nxi_arr_Mv_{M_v:.3f}at_z={self.z:.2f}_A{self.A2byA1}_k{self.kMpc_trans}_alpha{self.alpha}_beta{self.beta}.npy')
        except FileNotFoundError:
            os.makedirs(f'.Nxi_dz_Interp_init/NxiAtz{self.z:.2f}', exist_ok=True)
            M_min = self.M_J
            M_max = self.M_min
            Mh_interp = np.linspace(np.log10(M_min), np.log10(M_max), 1000)  # log10(M)
            deltaR_grid, m_grid = np.meshgrid(self.deltaV_interp, Mh_interp, indexing='ij')
            mh = self.dNxi_dz(m=10**m_grid, deltaR=deltaR_grid, Mv=M_v, z=z)
            mh[mh > 0] = 0
            integrand = mh * 10**m_grid * np.log(10)  # dN/dlogM -> dN/dM
            # Integrate using Simpson's rule
            Nxi_arr_save = simpson(integrand, x=Mh_interp, axis=1)
            np.save(f'.Nxi_dz_Interp_init/NxiAtz{self.z:.2f}/Nxi_arr_Mv_{M_v:.3f}at_z={self.z:.2f}_A{self.A2byA1}_k{self.kMpc_trans}_alpha{self.alpha}_beta{self.beta}.npy', Nxi_arr_save)
            Nxi_arr = np.load(f'.Nxi_dz_Interp_init/NxiAtz{self.z:.2f}/Nxi_arr_Mv_{M_v:.3f}at_z={self.z:.2f}_A{self.A2byA1}_k{self.kMpc_trans}_alpha{self.alpha}_beta{self.beta}.npy')
        Simpson_Interp = interp1d(self.deltaV_interp, Nxi_arr, kind='cubic')
        return Simpson_Interp(delta_R)
    
    ### Minihalo Initialization
    def Nxi_Pure(self,Mv,deltaR):
        m_H = (m_p.to(u.M_sun)).value #M_sun
        def Nxi_Pure_diff(m,Mv,deltaR):
            return self.Xim(m,self.z)*m*self.dndmeps(m,Mv,deltaR,self.z)/ m_H * self.cosmo.omegab / self.cosmo.omegam
        mslice = np.logspace(np.log10(self.M_J), np.log10(self.M_min), 12)
        ans = np.zeros_like(deltaR)
        for i in range(len(mslice)-1):
            ans += quad_vec(Nxi_Pure_diff, mslice[i], mslice[i+1],args=(Mv,deltaR), epsrel=1e-6)[0]
        return ans
    
    def Nxi_interp(self, Mv, deltaR):
        try:
            Nxi_arr = np.load(f'.Nxi_Interp_init/NxiAtz{self.z:.2f}/Nxi_arr_Mv_{Mv:.3f}at_z={self.z:.2f}_A{self.A2byA1}_k{self.kMpc_trans}_alpha{self.alpha}_beta{self.beta}.npy')
        except FileNotFoundError:
            os.makedirs(f'.Nxi_Interp_init/NxiAtz{self.z:.2f}', exist_ok=True)
            nxi_pure = self.Nxi_Pure(Mv, self.deltaV_interp)
            np.save(f'.Nxi_Interp_init/NxiAtz{self.z:.2f}/Nxi_arr_Mv_{Mv:.3f}at_z={self.z:.2f}_A{self.A2byA1}_k{self.kMpc_trans}_alpha{self.alpha}_beta{self.beta}.npy', nxi_pure)
            Nxi_arr = np.load(f'.Nxi_Interp_init/NxiAtz{self.z:.2f}/Nxi_arr_Mv_{Mv:.3f}at_z={self.z:.2f}_A{self.A2byA1}_k{self.kMpc_trans}_alpha{self.alpha}_beta{self.beta}.npy')
        Nxi_interp_Mv = interp1d(self.deltaV_interp, Nxi_arr, kind='cubic')
        return Nxi_interp_Mv(deltaR) 
    
    ### Minihaolo ST_over_PS
    def Nxi_ST(self,z):
        m_H = (m_p.to(u.M_sun)).value #M_sun
        Min = self.cosmo.M_Jeans(z)
        Max = self.cosmo.M_vir(0.61,1e4,z)
        def Nxi_ST_diff(m):
            return (self.Xim(m,z)*1 / m_H * self.cosmo.omegab / self.cosmo.omegam * m * self.cosmo.dndmst(m, z))
        mslice = np.logspace(np.log10(Min), np.log10(Max), 100)
        ans = 0
        for i in range(len(mslice)-1):
            ans += quad(Nxi_ST_diff, mslice[i], mslice[i+1], epsrel=1e-5)[0]
        return ans
    
    def dST_dz(self,m,z):
        return (self.cosmo.dndmst(m,z+0.001*z) - self.cosmo.dndmst(m,z-0.001*z)) / (0.002*z)

    def dNxi_dz_ST(self,m,z):
        m_H = (m_p.to(u.M_sun)).value #M_sun
        omega_b = self.cosmo.omegab
        omega_m = self.cosmo.omegam
        return self.Xim(m,z)*m*self.dST_dz(m,z)/ m_H * omega_b / omega_m

    def dNxi_ST_Simpson(self,z):
        Min = self.cosmo.M_Jeans(z)
        Max = self.cosmo.M_vir(0.61,1e4,z)
        ms = np.linspace(np.log10(Min), np.log10(Max), 1000)
        dNxist = self.dNxi_dz_ST(10**ms,z)
        dNxist[dNxist > 0] = 0
        integrand = dNxist * 10**ms * np.log(10)
        arr = simpson(integrand, ms, axis=0)
        return arr

    def Nxi_ST_AtZ(self,z):
        dz = -0.1
        zlin = np.arange(20, z + dz/2, dz)
        dNxi_ST = self.dNxi_ST_Simpson(zlin)
        Nxi_st = abs(np.trapezoid(dNxi_ST, zlin)) + self.Nxi_ST(20)
        return Nxi_st



"""
From the matter power spectrum to sigma(M) , dsigma/dM and Massfunction.
"""
from dataclasses import dataclass
import os
import numpy as np
import camb
from scipy.interpolate import interp1d
from scipy.integrate import quad,quad_vec
from filelock import FileLock
import astropy.units as u
from scipy.integrate import simpson
import massfunc as mf 


@dataclass
class CosmologyParams:
    """Set Cosmological parameters."""
    h: float = 0.674
    omegam: float = 0.315

    def __post_init__(self):
        self.omegab = 0.0224 * self.h**-2
        self.omegalam = 1 - self.omegam
        self.rhocrit = 2.775366e11 * self.h**2 * u.Msun / u.Mpc**3 
        rhoc = self.rhocrit.value
        self.rhom = rhoc * self.omegam
        self.H0u = 100 * self.h * (u.km * u.s**-1 * u.Mpc**-1)
        self.mHu = 1.6726e-27 * u.kg                        # the mass of a hydrogen Unit: kg
        self.X = 0.75                                       # the mass fraction of hydrogen
        self.nHu = self.rhocrit.to(u.kg/u.cm**3) * self.omegab * self.X / self.mHu    # hydrogen density
        self.nH = self.nHu.value
        self.omegak = 0.0   
        self.omegar = 0.0

class SteepPowerCalculator:

    #here As, ns are from Planck constraints, you should not change them much
    #alpha, beta, A2byA1, kMpc_trans are model parameters; alpha in [-2, 2], beta in [-pi, pi], A2byA1 in (0, 1), kMpc_trans is where you want the boost of power to start.
    # To get the standard case without power boost, you can set A2byA1 = 1 or kMpc_trans = + infinity.
    def __init__(self, As=2.1e-9, ns = 0.965, A2byA1 = 0.1, kMpc_trans = 1200, alpha=2.0, beta=0.0):
        self.alpha = alpha
        self.beta = beta
        self.c_alpha = np.cosh(alpha)
        self.s_alpha = np.sinh(alpha)
        self.c_beta = np.cos(beta)
        self.s_beta = np.sin(beta)
        self.As = As
        self.ns = ns
        self.A2byA1 = A2byA1
        self.kMpc_trans = kMpc_trans
        self.kMpc_pivot = 0.05 ##where As is defined
        eps = (1.-ns)/6.  #\epsilon        
        self.boost = (self.c_alpha-self.c_beta*self.s_alpha)**2 + (self.s_beta*self.s_alpha)**2

        self.HbyMp = np.pi*np.sqrt(As*eps*8./self.boost)  #H/M_p

        self.A1byMp3 = np.sqrt(eps*18.)*self.HbyMp**2
        self.A2byMp3 = self.A1byMp3*A2byA1

        self.norm = (3.*self.HbyMp**3/(2.*np.pi*self.A2byMp3))**2
        self.coef  = 3.*(1. - self.A2byA1)
        self.c1 = (self.c_alpha-self.c_beta*self.s_alpha)
        self.c2 = self.s_beta*self.s_alpha
        self.c3 = (self.c_alpha+self.c_beta*self.s_alpha)
        self.mp_prepared = False
                


    def PrimordialPower(self, kMpc:list):
        k = kMpc/self.kMpc_trans
        ksq = k**2
        if(k < 0.03): 
            kjy = -1./3.+ ksq*( (2./35.)*ksq -2./15. )
            kjj = k*ksq/9.*(1.-ksq/5.)
        else:
            kjy = ((k**2-1.)*np.sin(2*k)+2*k*np.cos(2*k))/(2.*k**3)
            kjj = (k*np.cos(k)-np.sin(k))**2/k**3
        return self.norm*(kMpc/self.kMpc_pivot)**(self.ns-1.)*((self.c1*(1.+self.coef*kjy) + self.c2*self.coef*kjj)**2 + (self.c2*(1.+self.coef*kjy) + self.c3*self.coef*kjj)**2)

   # Transfer function from BBKS to primordial power
    def BBKS_trans(self, x:float):
        return  np.log(1.+0.171*x)**2 *x**(self.ns-2.) / np.sqrt( 1.+ x*(0.284+x*(1.3924+x*(0.0635212+x*0.057648))))

    def Prepare_MatterPower(self, H0 = 67.5, ombh2 = 0.022, omch2 = 0.12, kMpc_max=10., zmax = 100.):
        nz = 100 #number of steps to use for the radial/redshift integration
        kmax=min(kMpc_max, 30.)  #kmax to use
        pars = camb.CAMBparams()
        pars.set_cosmology(H0=H0, ombh2=ombh2, omch2=omch2)
        pars.InitPower.set_params(As = self.As, ns=self.ns)
        results= camb.get_background(pars)
        self.mpi =  camb.get_matter_power_interpolator(pars, nonlinear=False, hubble_units=False, k_hunit=False, kmax=kmax, zmax=zmax)
        self.k_eq = 0.073*(omch2+ombh2)
        self.kmax = kmax
        self.norm_edge = self.BBKS_trans(self.kmax/self.k_eq)
        self.mp_prepared = True



    def MatterPower(self, kMpc:list, z): #kMpc can be an ordered array; z is the redshift
        if(not self.mp_prepared):
            self.Prepare_MatterPower()
        if(isinstance(kMpc, float)):
            if(kMpc <= self.kmax):
                pkraw = self.mpi.P(z, kMpc)
            else:
                pkraw = self.BBKS_trans(kMpc/self.k_eq)/self.norm_edge * self.mpi.P(z, self.kmax)
            return pkraw * (self.PrimordialPower(kMpc)/(self.As*(kMpc/self.kMpc_pivot)**(self.ns-1.)))
        else:
            nk = len(kMpc)
            Pk = np.empty(nk)
            for i in range(nk):
                if(kMpc[i] <= self.kmax):
                    pkraw = self.mpi.P(z, kMpc[i])
                else:
                    pkraw = self.BBKS_trans(kMpc[i]/self.k_eq)/self.norm_edge*self.mpi.P(z, self.kmax)
                Pk[i] = pkraw * (self.PrimordialPower(kMpc[i])/(self.As*(kMpc[i]/self.kMpc_pivot)**(self.ns-1.)))
            return Pk

class MassFunctions(CosmologyParams):
    def __init__(self, A2byA1=0.1, kMpc_trans=2e2, alpha=2.0, beta=0.0,h=0.674, omegam=0.315):
        super().__init__(h=h, omegam=omegam)
        self.cosmo = mf.SFRD(h=h,omegam=omegam)
        self.spc = SteepPowerCalculator(As = 2.1e-9, ns = 0.965, A2byA1 = A2byA1, kMpc_trans = kMpc_trans, alpha = alpha, beta = beta) 
        self.A2byA1, self.kMpc_trans, self.alpha, self.beta = A2byA1, kMpc_trans, alpha, beta
        self.dsigma2_dm_interpolation_completed = False
        self.pk_interpolation_completed = False
        self.sigma2_interpolation_completed = False
        self.m_interp_range = np.logspace(0.0,18.0,1000)
        self.ps_interp_range = np.logspace(-4.0,6.0,50000)
    

    def PowerSpectrum_Interp_Set(self):
        filename = f'.ps_init_out/Pk_interp_A{self.A2byA1}_K{self.kMpc_trans}_Alpha{self.alpha}_Beta{self.beta}.npy'
        lockfile = filename + '.lock'
        try:
            Pk_arr = np.load(filename)
        except FileNotFoundError:
            os.makedirs('.ps_init_out', exist_ok=True)
            with FileLock(lockfile):
                if not os.path.exists(filename):
                    self.spc.Prepare_MatterPower(H0=67.4, ombh2=0.022, omch2=0.12, kMpc_max=self.ps_interp_range[-1])
                    Pk = self.spc.MatterPower(self.ps_interp_range, z=0.0)
                    np.save(filename, Pk)
                Pk_arr = np.load(filename)
        self.pk0_interpolation = interp1d(np.log10(self.ps_interp_range), np.log10(Pk_arr), kind='cubic')
        self.pk_interpolation_completed = True

    def Sigma2_Interp_Set(self):
        filename = f'.ps_init_out/Sigma2_interp_A{self.A2byA1}_K{self.kMpc_trans}_Alpha{self.alpha}_Beta{self.beta}.npy'
        lockfile = filename + '.lock'
        try:
            sig_arr = np.load(filename)
        except FileNotFoundError:
            os.makedirs('.ps_init_out', exist_ok=True)
            with FileLock(lockfile):
                if not os.path.exists(filename):
                    sig = self.sigma2(self.m_interp_range)
                    np.save(filename, sig)
                sig_arr = np.load(filename)
        self.sigma2m_interpolation = interp1d(np.log10(self.m_interp_range), np.log10(sig_arr), kind='cubic')
        self.sigma2_interpolation_completed = True

    def Dsigma2dm_Interp_Set(self):
        filename = f'.ps_init_out/Dsigma2_dm_interp_A{self.A2byA1}_K{self.kMpc_trans}_Alpha{self.alpha}_Beta{self.beta}.npy'
        lockfile = filename + '.lock'
        try:
            dsig2dm_arr = np.load(filename)
        except FileNotFoundError:
            os.makedirs('.ps_init_out', exist_ok=True)
            with FileLock(lockfile):
                if not os.path.exists(filename):
                    dsig2dm = np.log10(-self.dsigma2_dm(self.m_interp_range))
                    np.save(filename, dsig2dm)
                dsig2dm_arr = np.load(filename)
        self.dsigma2_dm_interpolation = interp1d(np.log10(self.m_interp_range), dsig2dm_arr, kind='cubic')
        self.dsigma2_dm_interpolation_completed = True

    def ps_interp(self, kMpc):
        if not self.pk_interpolation_completed:
            self.PowerSpectrum_Interp_Set()
        return 10**self.pk0_interpolation(np.log10(kMpc))

    def sigma2_interp(self,M):
        if not self.sigma2_interpolation_completed:
            self.Sigma2_Interp_Set()
        return 10**self.sigma2m_interpolation(np.log10(M))
    
    def dsigma2_dm_interp(self,M):
        if not self.dsigma2_dm_interpolation_completed:
            self.Dsigma2dm_Interp_Set()
        return -10 ** self.dsigma2_dm_interpolation(np.log10(M))

    def dsigma2_dk(self,M,k):
        r=(3.0*M/(4.0*np.pi*self.rhom))**(1./3.) 
        x=r*k
        w=3.0*(np.sin(x)-x*np.cos(x))/x**3.0
        return 4.0*np.pi*k**2.0/(2.0*np.pi)**3.0 * self.ps_interp(k) * w*w

    def dsigma2_dlnk(self,lnk,M):
        k=np.exp(lnk)
        return k*self.dsigma2_dk(M,k)
    
    def sigma2(self,M):
        precision = 1e-6
        k_grid_1 = np.logspace(-4,0.9,20)
        k_grid_2 = np.logspace(1,5.3,200)
        k_grid = np.concatenate([k_grid_1, k_grid_2])
        int_ans = np.zeros_like(M)
        for i in range(len(k_grid)-1):
            int_ans += quad_vec(self.dsigma2_dlnk,np.log(k_grid[i]),np.log(k_grid[i+1]), args=(M,),epsrel=precision,limit=200)[0]
        return int_ans

    def sigma2_tarpz(self, M):
        k_grid_1 = np.logspace(-4, 0.999, 500)  # 增加点数提高精度
        k_grid_2 = np.logspace(1, 5.3, 2000)
        k_grid = np.concatenate([k_grid_1, k_grid_2])
        lnk_grid = np.log(k_grid)
        M_mesh, lnk_mesh = np.meshgrid(M, lnk_grid, indexing='ij')
        integrand_values = self.dsigma2_dlnk(lnk_mesh.T, M_mesh.T).T
        return np.trapezoid(integrand_values, x=lnk_grid, axis=1)

    def sigma2_simpson(self, M):
        k_grid_1 = np.logspace(-4, 0.999, 501)  
        k_grid_2 = np.logspace(1, 5.3, 2001)
        k_grid = np.concatenate([k_grid_1, k_grid_2])
        lnk_grid = np.log(k_grid)
        M_mesh, lnk_mesh = np.meshgrid(M, lnk_grid, indexing='ij')
        integrand_values = self.dsigma2_dlnk(lnk_mesh.T, M_mesh.T).T
        return simpson(integrand_values, x=lnk_grid, axis=1)

    def dsigma2_dlnk_dm(self,lnk,M):
        k=np.exp(lnk)
        r=(3.0*M/(4.0*np.pi*self.rhom))**(1./3.)
        x=r*k
        w=3.0*(np.sin(x)-x*np.cos(x))/(x)**3.0
        dw_dx=(9.0*x*np.cos(x)+np.sin(x)*(3.0*x**2.0-9.0))/(x)**4.0
        dw_dm=dw_dx*k/(4.0*np.pi*r**2.0*self.rhom)
        return 4.0*np.pi*k**3.0/(2.0*np.pi)**3.0*self.ps_interp(k)*2.0*w*dw_dm

    def dsigma2_dm(self,M):
        precision = 1e-6
        k_grid_1 = np.logspace(-4,0.999,20)

        k_grid_2 = np.logspace(1,5.3,200)
        k_grid = np.concatenate([k_grid_1, k_grid_2])
        int_ans = np.zeros_like(M)
        for i in range(len(k_grid)-1):
            int_ans += quad_vec(self.dsigma2_dlnk_dm,np.log(k_grid[i]),np.log(k_grid[i+1]), args=(M,),epsrel=precision,limit=200)[0]
        return int_ans

    def dsigma2_dm_simpson(self, M):
        k_grid_1 = np.logspace(-4, 0.999, 501)  # 奇数点数
        k_grid_2 = np.logspace(1, 5.3, 2001)
        k_grid = np.concatenate([k_grid_1, k_grid_2])
        lnk_grid = np.log(k_grid)
        
        # 向量化计算
        M_mesh, lnk_mesh = np.meshgrid(M, lnk_grid, indexing='ij')
        integrand_values = self.dsigma2_dlnk_dm(lnk_mesh.T, M_mesh.T).T
        
        return simpson(integrand_values, x=lnk_grid, axis=1)    

    def dsigma2_dm_trapz(self, M):
        k_grid_1 = np.logspace(-4, 0.999, 1000)  # 任意点数
        k_grid_2 = np.logspace(1, 5.3, 5000)
        k_grid = np.concatenate([k_grid_1, k_grid_2])
        lnk_grid = np.log(k_grid)
        
        # 向量化计算
        M_mesh, lnk_mesh = np.meshgrid(M, lnk_grid, indexing='ij')
        integrand_values = self.dsigma2_dlnk_dm(lnk_mesh.T, M_mesh.T).T
        
        return np.trapezoid(integrand_values, x=lnk_grid, axis=1)
    
    def deltac(self, z):
        return self.cosmo.deltac(z)
    
    def dndmps(self, m, z):
        sigm = np.sqrt(self.sigma2_interp(m))
        dsig_dm = abs(self.dsigma2_dm_interp(m)) / (2.0 * sigm)
        return np.sqrt(2.0 / np.pi) * self.rhom / m * self.deltac(z) / sigm**2 * dsig_dm * np.exp(-self.deltac(z) ** 2 / (2 * sigm ** 2))

    def dndmst(self,M,z):
        Ast=0.353
        ast2=0.73
        Pst=0.175
    #   ST parameters from Jenkins et al. 2001,0005260
        sigma = np.sqrt(self.sigma2_interp(M))
        nu=self.deltac(z)/sigma
        nup=np.sqrt(ast2)*nu
        dsigmadm=self.dsigma2_dm_interp(M)/(2*sigma)
        return -self.rhom/M*(dsigmadm/sigma)*(2*Ast)*(1+1.0/(nup)**(2*Pst))*(nup**2/(2*np.pi))**(1./2.)*np.exp(-nup**2/2)

    def fcoll_st(self, z,Mmin, Mmax=1e16,n=2000):
        logm_max = np.log10(Mmax)
        logm_min = np.log10(Mmin)
        def diff(m,z):
            return self.dndmst(m,z)*m
        x = np.linspace(logm_min, logm_max, n)
        y = diff(10**x, z)*10**x*np.log(10)
        integral = np.trapezoid(y, x)        
        return integral / self.rhom

    def delta_L(self, deltar,z):
        return (1.68647 - 1.35 / (1 + deltar) ** (2 / 3) - 1.12431 / (1 + deltar) ** (1 / 2) + 0.78785 / (1 + deltar) ** (0.58661)) / self.cosmo.Dz(z)

    def dndmeps(self, M, Mr, deltar, z):
        delta_L = self.delta_L(deltar,z)
        sig1 = self.sigma2_interp(M) - self.sigma2_interp(Mr)
        del1 = self.cosmo.deltac(z) - delta_L
        return self.rhom * (1 + deltar) / M / np.sqrt(2 * np.pi) * abs(self.dsigma2_dm_interp(M)) * del1 / sig1 ** (3 / 2) * np.exp(-del1 ** 2 / (2 * sig1))
    
    def M_Jeans(self, z):
        return 5.73e3*(self.omegam*self.h**2/0.15)**(-1/2) * (self.omegab*self.h**2/0.0224)**(-3/5) * ((1+z)/10)**(3/2)

    def Delta_cc(self,z):
        d=self.cosmo.omegam_z(z)-1.0
        return 18*np.pi**2+82.0*d-39.0*d**2

    def M_vir(self,mu,Tvir,z):
        a1=(self.cosmo.omegam_z(z)*self.Delta_cc(z)/(18*np.pi**2))**(-1.0/3.0)
        a2=a1*(mu/0.6)**(-1.0)*((1.0+z)/10)**(-1.0)/1.98e4*Tvir
        return a2**(3.0/2.0)*1e8/self.h


from .powerspec import MassFunctions
import numpy as np
import cupy as cp
import astropy.units as u
from cosfunc import n_H,dtdz
from .special import qion_sb99,interp1d_gpu,xim,fstar

class Ion:

    def __init__(self,z,fesc=0.2,A2byA1=0.1,ktrans=200,alpha=2.0,beta=0.0):
        self.cosmo = MassFunctions(A2byA1=A2byA1,kMpc_trans=ktrans,alpha=alpha,beta=beta)
        self.z = z
        self.mH = (self.cosmo.mHu.to(u.M_sun)).value #M_sun
        self.ob0 = self.cosmo.omegab
        self.om0 = self.cosmo.omegam
        self.nH = self.cosmo.nH  #cm^-3
        self.fesc = fesc
        self.deltaV_interp = np.linspace(-0.951, 2, 100)  # delta_V
        self.deltav_gpu = cp.asarray(self.deltaV_interp)
        self.M_min = self.cosmo.M_vir(0.61, 1e4, z)  # M_sun
        self.M_max = self.cosmo.M_vir(0.61, 2e8, z)
        self.M_J = self.cosmo.M_Jeans(z)  # M_sun
        self.qion = qion_sb99(z)  # ionizing photon per baryon in stars

    ### Source term
    def nion_interp(self,Mv:float,deltav:cp.ndarray)->cp.ndarray:
        def diff(m,Mv,deltaR):
            return fstar(m)*m*self.cosmo.dndmeps(m,Mv,deltaR,self.z)/ self.mH * self.ob0 / self.om0
        return self._nion_trapez(diff,Mv,deltav)*self.qion*self.fesc
    
    def nion_st(self,z):
        def diff(m):
            return fstar(m)*m*self.cosmo.dndmst(m,z)/ self.mH * self.ob0 / self.om0
        x = np.linspace(np.log10(self.M_min),np.log10(self.M_max),1000)
        y = diff(10**x)*10**x*np.log(10)
        return np.trapezoid(y,x)*self.qion*self.fesc
    
    def nion_ps(self,z):
        def diff(m):
            return fstar(m)*m*self.cosmo.dndmps(m,z)/ self.mH * self.ob0 / self.om0
        x = np.linspace(np.log10(self.M_min),np.log10(self.M_max),1000)
        y = diff(10**x)*10**x*np.log(10)
        return np.trapezoid(y,x)*self.qion*self.fesc
    
    ### Minihalo term
    def nxi_interp(self, Mv, deltav):
        def diff(m, Mv, deltav):
            return xim(m,self.z) * m * self.cosmo.dndmeps(m, Mv, deltav, self.z) / self.mH * self.ob0 / self.om0
        return self._nxi_trapez(diff, Mv, deltav)

    def nxi_st(self, z):
        def diff(m):
            return xim(m,z) * m * self.cosmo.dndmst(m,z) / self.mH * self.ob0 / self.om0
        x = np.linspace(np.log10(self.M_J),np.log10(self.M_min),1000)
        y = diff(10**x)*10**x*np.log(10)
        return np.trapezoid(y,x)

    ### IGM term
    def n_HI(self,deltaV):
        return ((1/u.cm**3).to(u.Mpc**-3)).value*n_H(deltaV) ### coming number density of hydrogen atoms in cm^-3
    
    def CHII(self,z):
        return 2.9*((1+z)/6)**-1.1

    def dnrec_dz_path(self,deltaV,xHII_Field:cp.ndarray)->cp.ndarray:
        x_HE = 1.08
        CIGM = self.CHII(self.z)
        # CIGM = 3.0
        nh = self.n_HI(deltaV)*(u.Mpc**-3).to(u.cm**-3)
        Q_HII = xHII_Field
        alpha_A = 4.2e-13 #cm**3/s
        # alpha_B = 2.59e-13  # cm^3/s at 10^4 K
        differential_trans = dtdz(self.z).to(u.s).value
        return CIGM*x_HE*alpha_A*nh*Q_HII*(1+self.z)**3 * differential_trans
    #### medium function




    #### method functions
    def _nxi_trapez(self, diff_func, Mv, deltav):
        x = np.linspace(np.log10(self.M_J), np.log10(self.M_min), 1000)
        deltaV_grid, m_grid = np.meshgrid(self.deltaV_interp, x, indexing='ij')
        m_vals = 10**m_grid
        y = diff_func(m_vals, Mv, deltaV_grid) * m_vals * np.log(10)
        integrand = np.trapezoid(y, x, axis=1)
        integrand = cp.asarray(integrand)
        return interp1d_gpu(self.deltav_gpu, integrand, deltav)
    
    def _nion_trapez(self, diff_func, Mv,deltav):
        Mh_max = 0.9*min(self.M_max,Mv)
        x = np.linspace(np.log10(self.M_min), np.log10(Mh_max), 1000)
        deltaV_grid, m_grid = np.meshgrid(self.deltaV_interp, x, indexing='ij')
        m_vals = 10**m_grid
        y = diff_func(m_vals, Mv, deltaV_grid) * m_vals * np.log(10)
        integrand = np.trapezoid(y, x, axis=1)
        integrand = cp.asarray(integrand)
        return interp1d_gpu(self.deltav_gpu, integrand, deltav)
    
    


from .powerspec import MassFunctions
import numpy as np
from scipy.interpolate import interp1d
import astropy.units as u
from cosfunc import n_H,dtdz
from astropy.constants import m_p
from .special import qion_sb99


class Ion:

    def __init__(self,z,fesc=0.2,A2byA1=0.1,ktrans=200,alpha=2.0,beta=0.0):
        self.cosmo = MassFunctions(A2byA1=A2byA1,kMpc_trans=ktrans,alpha=alpha,beta=beta)
        self.z = z
        self.mH = (self.cosmo.mHu.to(u.M_sun)).value #M_sun
        self.ob0 = self.cosmo.omegab
        self.om0 = self.cosmo.omegam
        self.nH = self.cosmo.nH  #cm^-3
        self.fesc = fesc
        self.deltaV_interp = np.linspace(-0.999, 2, 100)  # delta_V
        self.M_min = self.cosmo.M_vir(0.61, 1e4, z)  # M_sun
        self.M_max = self.cosmo.M_vir(0.61, 2e8, z)
        self.M_J = self.cosmo.M_Jeans(z)  # M_sun
        self.qion = qion_sb99(z)  # ionizing photon per baryon in stars

    ### Source term
    def nion_interp(self,Mv:float,deltaV:np.ndarray)->np.ndarray:
        def diff(m,Mv,deltaR):
            return self.fstar(m)*m*self.cosmo.dndmeps(m,Mv,deltaR,self.z)/ self.mH * self.ob0 / self.om0
        return self._nion_trapez(diff,Mv,deltaV)*self.qion*self.fesc
    
    def nion_st(self,z):
        def diff(m):
            return self.fstar(m)*m*self.cosmo.dndmst(m,z)/ self.mH * self.ob0 / self.om0
        x = np.linspace(np.log10(self.M_min),np.log10(self.M_max),100000)
        y = diff(10**x)*10**x*np.log(10)
        return np.trapezoid(y,x)*self.qion*self.fesc
    
    def nion_ps(self,z):
        def diff(m):
            return self.fstar(m)*m*self.cosmo.dndmps(m,z)/ self.mH * self.ob0 / self.om0
        x = np.linspace(np.log10(self.M_min),np.log10(self.M_max),100000)
        y = diff(10**x)*10**x*np.log(10)
        return np.trapezoid(y,x)*self.qion*self.fesc
    
    ### Minihalo term
    def eps_dz(self,M,Mv,deltaV,z):
        return (self.cosmo.dndmeps(M,Mv,deltaV,z+0.001*z) - self.cosmo.dndmeps(M,Mv,deltaV,z-0.001*z)) / (0.002*z)
    
    def nxi_dz_interp(self, Mv, deltaV):
        def diff(m, Mv, deltaV):
            return self.xim(m,self.z) * m * self.eps_dz(m, Mv, deltaV, self.z) / self.mH * self.ob0 / self.om0
        return self._nxi_trapez(diff, Mv, deltaV,zero_cut=True)
        
    def nxi_interp(self, Mv, deltaV):
        def diff(m, Mv, deltaV):
            return self.xim(m,self.z) * m * self.cosmo.dndmeps(m, Mv, deltaV, self.z) / self.mH * self.ob0 / self.om0
        return self._nxi_trapez(diff, Mv, deltaV)
    
    ### Minihaolo ST_over_PS
    def nxi_st_ini(self,z):
        def diff(m):
            return (self.xim(m,z)*1 / self.mH * self.ob0 / self.om0 * m * self.cosmo.dndmst(m, z))
        x = np.linspace(np.log10(self.cosmo.M_Jeans(z)), np.log10(self.cosmo.M_vir(0.61,1e4,z)), 1000)
        y = diff(10**x)*10**x *np.log(10)
        return np.trapezoid(y,x)
    
    def st_dz(self,m,z):
        return (self.cosmo.dndmst(m,z+0.01) - self.cosmo.dndmst(m,z-0.01)) / 0.02

    def nxi_dz_st(self, z: np.ndarray):
        z = np.asarray(z)
        z_shape = z.shape
        z = z.ravel()  # 展平，便于广播，最后再 reshape

        # 固定大质量网格（覆盖所有可能 z 的质量范围）
        log_m_min_global = 3.0   # 1e3 M☉
        log_m_max_global = 10.0  # 1e10 M☉
        x = np.linspace(log_m_min_global, log_m_max_global, 1000)  # (1000,)
        m = 10**x  # (1000,)

        # 为每个 z 计算实际质量边界
        M_J_z = self.cosmo.M_Jeans(z)      # (N,)
        M_min_z = self.cosmo.M_vir(0.61, 1e4, z)  # (N,)

        # 扩展维度以便广播
        m_grid = m[:, None]      # (1000, 1)
        z_grid = z[None, :]      # (1, N)
        M_J_grid = M_J_z[None, :]    # (1, N)
        M_min_grid = M_min_z[None, :] # (1, N)

        # 计算被积函数（先不管质量范围）
        diff_val = self.xim(m_grid, z_grid) * m_grid * self.st_dz(m_grid, z_grid) / self.mH * self.ob0 / self.om0
        y = diff_val * m_grid * np.log(10)  # (1000, N)

        # 应用质量范围 mask：只在 M_J(z) <= m <= M_min(z) 内有效，其余设为 0
        mask = (m_grid >= M_J_grid) & (m_grid <= M_min_grid)  # (1000, N)
        y = np.where(mask, y, 0.0)

        # 只保留负贡献（可选，但保险）
        y = np.minimum(y, 0)

        # 沿质量轴积分 → (N,)
        result = np.trapezoid(y, x, axis=0)

        return result.reshape(z_shape)

    def nxi_st(self,z):
        zlin = np.arange(20.1, z-0.05, -0.1)
        step = self.nxi_dz_st(zlin)
        return abs(np.trapezoid(step, zlin)) + self.nxi_st_ini(20.1)


    ### IGM term
    def n_HI(self,deltaV):
        return n_H(deltaV).to(u.Mpc**-3).value  ### coming number density of hydrogen atoms in cm^-3
    
    def CHII(self,z):
        return 2.9*((1+z)/6)**-1.1

    def dnrec_dz_path(self,deltaV,xHII_Field:np.ndarray)->np.ndarray:
        x_HE = 1.08
        CIGM = self.CHII(self.z)
        # CIGM = 3.0
        nh = self.n_HI(deltaV)*(u.Mpc**-3).to(u.cm**-3)
        Q_HII = xHII_Field
        alpha_A = 4.2e-13 #cm**3/s
        # alpha_B = 2.59e-13  # cm^3/s at 10^4 K
        differential_trans = dtdz(self.z).to(u.s).value
        return CIGM*x_HE*alpha_A*nh*Q_HII*(1+self.z)**3 * differential_trans
    

    #### medium function
    def xim(self,m,z):
        A,B,C,D,E,F,G = 4.4 , 0.334, 0.023, 0.199, -0.042, 0,1
        M7 = m/1e7
        F0 = 1.0
        return 1+A*M7**(B+C*np.log10(M7))*(F+G*(1+z)/10)* F0**(D+E*np.log10(F0))

    def fstar(self, Mh): # Donnan 25 
        eps0 = 0.16
        Mc = 10**11.7
        beta = 0.9
        gamma = 0.65
        return 2*eps0 * ((Mh / Mc)**-beta + (Mh / Mc)**gamma)**-1


    #### method functions
    def _nxi_trapez(self, diff_func, Mv, deltaV,zero_cut=False):
        x = np.linspace(np.log10(self.M_J), np.log10(self.M_min), 1000)
        deltaV_grid, m_grid = np.meshgrid(self.deltaV_interp, x, indexing='ij')
        m_vals = 10**m_grid
        y = diff_func(m_vals, Mv, deltaV_grid) * m_vals * np.log(10)
        if zero_cut:
            y = np.minimum(y,0)
        integrand = np.trapezoid(y, x, axis=1)
        interp_func = interp1d(self.deltaV_interp, integrand, kind='cubic', bounds_error=False, fill_value=0)
        return interp_func(deltaV)
    
    def _nion_trapez(self, diff_func, Mv, deltaV):
        Mh_max = 0.9*min(self.M_max,Mv)
        x = np.linspace(np.log10(self.M_min), np.log10(Mh_max), 1000)
        deltaV_grid, m_grid = np.meshgrid(self.deltaV_interp, x, indexing='ij')
        m_vals = 10**m_grid
        y = diff_func(m_vals, Mv, deltaV_grid) * m_vals * np.log(10)
        integrand = np.trapezoid(y, x, axis=1)
        interp_func = interp1d(self.deltaV_interp, integrand, kind='cubic', bounds_error=False, fill_value=0)
        return interp_func(deltaV)